package algorithms;

import types.procedures.*;
import types.structures.ITree;
import exceptions.NotFoundException;

// Represents binary tree-processing algorithms.
public class TreeAlgorithms {

	// Inserts obj into tree, assuming elements in tree must be
	// ordered left-to-right by order and that tree nodes must
	// be balanced enough to satisfy the given predicate.
	public static ITree insertBalanced(
			ITree tree, Object obj, 
			ICompare order, ISelect balanced ) {

		if ( tree.isLeaf() ) {
			return tree.hangLeft( obj, tree );
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			int comparison = order.compare( obj, data );
			
			if ( comparison < 0 ) {
				ITree left2 = insertBalanced( left, obj, order, balanced );
				ITree tree2 = left2.hangLeft( data, right );
				
				if ( balanced.select(tree2) ) {
					return tree2;
				} else {
					return rotateRight( tree2 );
				}
			} else if ( comparison > 0 ) {
				ITree right2 = insertBalanced( right, obj, order, balanced );
				ITree tree2 = right2.hangRight( left, data );
 
				if ( balanced.select(tree2) ) {
					return tree2;
				} else {
					return rotateLeft( tree2 );
				}
			} else {
				return left.hangLeft( obj, right );
			}
		}
	}

	/* Transforms tree:
	 *     o2            o1
	 *    /  \          /  \
	 *   o1  tC  into  tA  o2
	 *  /  \              /  \
	 * tA  tB            tB  tC
	 */
	public static ITree rotateRight( ITree tree ) {
		ITree left = tree.getLeft();
		Object o1 = left.getData();
		Object o2 = tree.getData();
		ITree tA = left.getLeft();
		ITree tB = left.getRight();
		ITree tC = tree.getRight();
		
		return tA.hangLeft( o1, tB.hangLeft( o2, tC ) );
	}
	
	/* Transforms tree:
	 *   o1                o2
	 *  /  \              /  \
	 * tA  o2    into    o1  tC
	 *    /  \          /  \
	 *   tB  tC        tA  tB
	 */
	public static ITree rotateLeft( ITree tree ) {
		ITree right = tree.getRight();
		Object o1 = tree.getData();
		Object o2 = right.getData();
		ITree tA = tree.getLeft();
		ITree tB = right.getLeft();
		ITree tC = right.getRight();

		return tA.hangLeft( o1, tB ).hangLeft( o2, tC );
	}

	// Produces data in sorted tree which compares as equal to obj,
	// or throws NotFoundException if no such data.
	public static Object find( ITree tree, Object obj, ICompare order )
	{
		if ( tree.isLeaf() ) {
			throw new NotFoundException();
		} else {
			Object data = tree.getData();
			int comparison = order.compare( obj, data );
			
			if ( comparison < 0 ) {
				return find( tree.getLeft(), obj, order );
			} else if ( comparison > 0 ) {
				return find( tree.getRight(), obj, order );
			} else {
				return data;
			}
		}
	}
	
	// Produces copy of tree with transform applied to data stored at nodes.
	public static ITree map( ITree tree, IConvert transform ) {
		if ( tree.isLeaf() ) {
			return tree;
		} else {
			Object data = transform.convert( tree.getData() );
			ITree left = map( tree.getLeft(), transform );
			ITree right = map( tree.getRight(), transform );
			
			return left.hangLeft( data, right );
		}
	}
	
	// Tells whether all data in tree satisfies predicate.
	public static boolean andMap( ITree tree, ISelect predicate ) {
		if ( tree.isLeaf() ) {
			return true;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return
				predicate.select( data ) &&
				andMap( left, predicate ) &&
				andMap( right, predicate );
		}
	}
	
	// Tells whether any data in tree satisfies predicate.
	public static boolean orMap( ITree tree, ISelect predicate ) {
		if ( tree.isLeaf() ) {
			return false;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return
				predicate.select( data ) ||
				orMap( left, predicate ) ||
				orMap( right, predicate );
		}
	}
	
	// Combines elements in tree in left->right inorder traversal.
	public static Object foldLeft( ITree tree, Object init, ICombine join ) {
		if ( tree.isLeaf() ) {
			return init;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return foldLeft( right, join.combine( data, foldLeft( left, init, join ) ), join );
		}
	}

	// Combines elements in tree in right->left inorder traversal.
	public static Object foldRight( ITree tree, Object init, ICombine join ) {
		if ( tree.isLeaf() ) {
			return init;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return foldRight( left, join.combine( data, foldRight( right, init, join ) ), join );
		}
	}
	
	// Combines elements in tree in left->right preorder traversal.
	public static Object foldLeftPreOrder( ITree tree, Object init, ICombine join ) {
		if ( tree.isLeaf() ) {
			return init;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return foldLeftPreOrder( right, foldLeftPreOrder( left, join.combine( data, init ), join ), join );
		}
	}

	// Combines elements in tree in left->right postorder traversal.
	public static Object foldLeftPostOrder( ITree tree, Object init, ICombine join ) {
		if ( tree.isLeaf() ) {
			return init;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return join.combine( data, foldLeftPostOrder( right, foldLeftPostOrder( left, init, join ), join ) );
		}
	}
	
	// Combines elements in tree in right->left preorder traversal.
	public static Object foldRightPreOrder( ITree tree, Object init, ICombine join ) {
		if ( tree.isLeaf() ) {
			return init;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return foldRightPreOrder( left, foldRightPreOrder( right, join.combine( data, init ), join ), join );
		}
	}

	// Combines elements in tree in right->left postorder traversal.
	public static Object foldRightPostOrder( ITree tree, Object init, ICombine join ) {
		if ( tree.isLeaf() ) {
			return init;
		} else {
			Object data = tree.getData();
			ITree left = tree.getLeft();
			ITree right = tree.getRight();
			
			return join.combine( data, foldRightPostOrder( left, foldRightPostOrder( right, init, join ), join ) );
		}
	}
	
}
