package structures.map;

import procedures.combinations.ListCons;
import procedures.conversions.EntryKey;
import procedures.conversions.EntryValue;
import procedures.orderings.EntryOrder;
import procedures.orderings.HashOrder;
import procedures.predicates.AVLInvariant;
import algorithms.ListAlgorithms;
import algorithms.TreeAlgorithms;
import structures.list.Empty;
import structures.tree.Leaf;
import types.procedures.ICompare;
import types.procedures.ISelect;
import types.structures.IMap;
import types.structures.ISame;
import types.structures.ITree;
import types.structures.IList;
import exceptions.NotFoundException;
import exceptions.HashCollisionException;

// Represents a mapping with unordered keys.
public class SimpleMap implements IMap {

	private static ISelect balanceInvariant = new AVLInvariant();
	private static ICompare entryOrder = new EntryOrder( new HashOrder() );
	
	private ITree tree;
	
	public SimpleMap() {
		this.tree = new Leaf();
	}
	
	private SimpleMap( ITree tree ) {
		this.tree = tree;
	}

	// Produces the key/value entry corresponding to key.
	// Throws HashCollisionException if distinct keys have the same hashcode.
	private Entry find( ISame key ) {
		Entry e =
			(Entry)TreeAlgorithms.find(
				this.tree, new Entry( key, null), entryOrder );
		if ( key.same((ISame)e.key) ) {
			return e;
		} else {
			throw new HashCollisionException();
		}
	}

	// Tells whether this has no key/value mappings.
	public boolean isEmpty() {
		return tree.isLeaf();
	}
	
	// Tells whether this maps key to a value.
	public boolean contains(ISame key) {
		try {
			this.find(key);
			return true;
		} catch( NotFoundException e ) {
			return false;
		}
	}

	// Produces the value which key is mapped to, or raises NotFoundException.
	// Raises HashCollisionException if key collides with an existing mapping.
	public Object lookup(ISame key) {
		return this.find(key).value;
	}

	// Produces a copy of this, mapping key to value.
	// Raises HashCollisionException if key collides with an existing mapping.
	public IMap extend(ISame key, Object value) {
		ITree tree =
			TreeAlgorithms.insertBalanced(
					this.tree, new Entry( key, value ), entryOrder, balanceInvariant );
		return new SimpleMap( tree );
	}

	// Produces a list of keys in this (in arbitrary order).
	public IList keys() {
		return ListAlgorithms.map( this.entries(), new EntryKey() );
	}

	// Produces a list of values in this (in order corresponding to the keys).
	public IList values() {
		return ListAlgorithms.map( this.entries(), new EntryValue() );
	}

	// Produces a list of key/value pairs in this (in arbitrary order).
	public IList entries() {
		return (IList)TreeAlgorithms.foldLeftPostOrder( this.tree, new Empty(), new ListCons() );
	}
	
	// Produces a string to display this.
	public String toString() {
		return "(map " + this.entries().toString() + ")";
	}

}
