package structures.set;

import exceptions.NotFoundException;
import exceptions.HashCollisionException;
import algorithms.TreeAlgorithms;
import procedures.combinations.ListCons;
import procedures.orderings.HashOrder;
import procedures.predicates.AVLInvariant;
import structures.list.Empty;
import structures.tree.Leaf;
import types.procedures.ICompare;
import types.procedures.ISelect;
import types.structures.IList;
import types.structures.ISame;
import types.structures.ISet;
import types.structures.ITree;

// Represents a set with unordered elements.
public class SimpleSet implements ISet {

	private static ICompare order = new HashOrder();
	private static ISelect balanceInvariant = new AVLInvariant();
	
	private ITree tree;
	
	public SimpleSet() {
		this.tree = new Leaf();
	}
	
	private SimpleSet( ITree tree ) {
		this.tree = tree;
	}
	
	// Tells whether this has no elements.
	public boolean isEmpty() {
		return this.tree.isLeaf();
	}

	// Tells whether element is in this.
	// Throws HashCollisionException if element has the same hashcode
	// as another distinct member of this.
	public boolean member(ISame element) {
		try {
			ISame found = (ISame)TreeAlgorithms.find( this.tree, element, order );
			if ( found.same(element) ) {
				return true;
			} else {
				throw new HashCollisionException();
			}
		}
		catch( NotFoundException e ) {
			return false;
		}
	}

	// Produces a copy of this containing element.
	// Throws HashCollisionException if element collides with a distinct member.
	public ISet insert(ISame element) {
		if ( this.member(element) ) {
			return this;
		} else {
			ITree tree =
				TreeAlgorithms.insertBalanced(
						this.tree, element, order, balanceInvariant );
			return new SimpleSet( tree );
		}
	}

	// Produces a list of all elements in this (in arbitrary order).
	public IList elements() {
		return (IList)TreeAlgorithms.foldLeftPreOrder( this.tree, new Empty(), new ListCons() );
	}
	
	// Produces a string to display this.
	public String toString() {
		return "(set " + this.elements().toString() + ")";
	}

}
